#! /usr/bin/env python

# kpwiki, a wiki-server in python
#
# Copyright (C) 2003  Kjell Enblom
#
#		    GNU GENERAL PUBLIC LICENSE
#		       Version 2, June 1991
#
# Copyright (C) 1989, 1991 Free Software Foundation, Inc.
#     59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# Everyone is permitted to copy and distribute verbatim copies
# of this license document, but changing it is not allowed.
#
#			    Preamble
#
#  The licenses for most software are designed to take away your
#freedom to share and change it.  By contrast, the GNU General Public
#License is intended to guarantee your freedom to share and change free
#software--to make sure the software is free for all its users.	This
#General Public License applies to most of the Free Software
#Foundation's software and to any other program whose authors commit to
#using it.  (Some other Free Software Foundation software is covered by
#the GNU Library General Public License instead.)  You can apply it to
#your programs, too.
#
#  When we speak of free software, we are referring to freedom, not
#price.	Our General Public Licenses are designed to make sure that you
#have the freedom to distribute copies of free software (and charge for
#this service if you wish), that you receive source code or can get it
#if you want it, that you can change the software or use pieces of it
#in new free programs; and that you know you can do these things.
#
#  To protect your rights, we need to make restrictions that forbid
#anyone to deny you these rights or to ask you to surrender the rights.
#These restrictions translate to certain responsibilities for you if you
#distribute copies of the software, or if you modify it.
#
#  For example, if you distribute copies of such a program, whether
#gratis or for a fee, you must give the recipients all the rights that
#you have.  You must make sure that they, too, receive or can get the
#source code.  And you must show them these terms so they know their
#rights.
#
#  We protect your rights with two steps: (1) copyright the software, and
#(2) offer you this license which gives you legal permission to copy,
#distribute and/or modify the software.
#
#  Also, for each author's protection and ours, we want to make certain
#that everyone understands that there is no warranty for this free
#software.  If the software is modified by someone else and passed on, we
#want its recipients to know that what they have is not the original, so
#that any problems introduced by others will not reflect on the original
#authors' reputations.
#
#
#  Finally, any free program is threatened constantly by software
#patents.  We wish to avoid the danger that redistributors of a free
#program will individually obtain patent licenses, in effect making the
#program proprietary.  To prevent this, we have made it clear that any
#patent must be licensed for everyone's free use or not licensed at all.
#
#  The precise terms and conditions for copying, distribution and
#modification follow.
#^L
#		    GNU GENERAL PUBLIC LICENSE
#   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
#
#  0. This License applies to any program or other work which contains
#a notice placed by the copyright holder saying it may be distributed
#under the terms of this General Public License.  The "Program", below,
#refers to any such program or work, and a "work based on the Program"
#means either the Program or any derivative work under copyright law:
#that is to say, a work containing the Program or a portion of it,
#either verbatim or with modifications and/or translated into another
#language.  (Hereinafter, translation is included without limitation in
#the term "modification".)  Each licensee is addressed as "you".
#
#Activities other than copying, distribution and modification are not
#covered by this License; they are outside its scope.  The act of
#running the Program is not restricted, and the output from the Program
#is covered only if its contents constitute a work based on the
#Program (independent of having been made by running the Program).
#Whether that is true depends on what the Program does.
#
#  1. You may copy and distribute verbatim copies of the Program's
#source code as you receive it, in any medium, provided that you
#conspicuously and appropriately publish on each copy an appropriate
#copyright notice and disclaimer of warranty; keep intact all the
#notices that refer to this License and to the absence of any warranty;
#and give any other recipients of the Program a copy of this License
#along with the Program.
#
#You may charge a fee for the physical act of transferring a copy, and
#you may at your option offer warranty protection in exchange for a fee.
#
#  2. You may modify your copy or copies of the Program or any portion
#of it, thus forming a work based on the Program, and copy and
#distribute such modifications or work under the terms of Section 1
#above, provided that you also meet all of these conditions:
#
#    a) You must cause the modified files to carry prominent notices
#    stating that you changed the files and the date of any change.
#
#    b) You must cause any work that you distribute or publish, that in
#    whole or in part contains or is derived from the Program or any
#    part thereof, to be licensed as a whole at no charge to all third
#    parties under the terms of this License.
#
#    c) If the modified program normally reads commands interactively
#    when run, you must cause it, when started running for such
#    interactive use in the most ordinary way, to print or display an
#    announcement including an appropriate copyright notice and a
#    notice that there is no warranty (or else, saying that you provide
#    a warranty) and that users may redistribute the program under
#    these conditions, and telling the user how to view a copy of this
#    License.  (Exception: if the Program itself is interactive but
#    does not normally print such an announcement, your work based on
#    the Program is not required to print an announcement.)
#^L
#These requirements apply to the modified work as a whole.  If
#identifiable sections of that work are not derived from the Program,
#and can be reasonably considered independent and separate works in
#themselves, then this License, and its terms, do not apply to those
#sections when you distribute them as separate works.  But when you
#distribute the same sections as part of a whole which is a work based
#on the Program, the distribution of the whole must be on the terms of
#this License, whose permissions for other licensees extend to the
#entire whole, and thus to each and every part regardless of who wrote it.
#
#Thus, it is not the intent of this section to claim rights or contest
#your rights to work written entirely by you; rather, the intent is to
#exercise the right to control the distribution of derivative or
#collective works based on the Program.
#
#In addition, mere aggregation of another work not based on the Program
#with the Program (or with a work based on the Program) on a volume of
#a storage or distribution medium does not bring the other work under
#the scope of this License.
#
#  3. You may copy and distribute the Program (or a work based on it,
#under Section 2) in object code or executable form under the terms of
#Sections 1 and 2 above provided that you also do one of the following:
#
#    a) Accompany it with the complete corresponding machine-readable
#    source code, which must be distributed under the terms of Sections
#    1 and 2 above on a medium customarily used for software interchange; or,
#
#    b) Accompany it with a written offer, valid for at least three
#    years, to give any third party, for a charge no more than your
#    cost of physically performing source distribution, a complete
#    machine-readable copy of the corresponding source code, to be
#    distributed under the terms of Sections 1 and 2 above on a medium
#    customarily used for software interchange; or,
#
#    c) Accompany it with the information you received as to the offer
#    to distribute corresponding source code.  (This alternative is
#    allowed only for noncommercial distribution and only if you
#    received the program in object code or executable form with such
#    an offer, in accord with Subsection b above.)
#
#The source code for a work means the preferred form of the work for
#making modifications to it.  For an executable work, complete source
#code means all the source code for all modules it contains, plus any
#associated interface definition files, plus the scripts used to
#control compilation and installation of the executable.  However, as a
#special exception, the source code distributed need not include
#anything that is normally distributed (in either source or binary
#form) with the major components (compiler, kernel, and so on) of the
#operating system on which the executable runs, unless that component
#itself accompanies the executable.
#
#If distribution of executable or object code is made by offering
#access to copy from a designated place, then offering equivalent
#access to copy the source code from the same place counts as
#distribution of the source code, even though third parties are not
#compelled to copy the source along with the object code.
#^L
#  4. You may not copy, modify, sublicense, or distribute the Program
#except as expressly provided under this License.  Any attempt
#otherwise to copy, modify, sublicense or distribute the Program is
#void, and will automatically terminate your rights under this License.
#However, parties who have received copies, or rights, from you under
#this License will not have their licenses terminated so long as such
#parties remain in full compliance.
#
#  5. You are not required to accept this License, since you have not
#signed it.  However, nothing else grants you permission to modify or
#distribute the Program or its derivative works.  These actions are
#prohibited by law if you do not accept this License.  Therefore, by
#modifying or distributing the Program (or any work based on the
#Program), you indicate your acceptance of this License to do so, and
#all its terms and conditions for copying, distributing or modifying
#the Program or works based on it.
#
#  6. Each time you redistribute the Program (or any work based on the
#Program), the recipient automatically receives a license from the
#original licensor to copy, distribute or modify the Program subject to
#these terms and conditions.  You may not impose any further
#restrictions on the recipients' exercise of the rights granted herein.
#You are not responsible for enforcing compliance by third parties to
#this License.
#
#  7. If, as a consequence of a court judgment or allegation of patent
#infringement or for any other reason (not limited to patent issues),
#conditions are imposed on you (whether by court order, agreement or
#otherwise) that contradict the conditions of this License, they do not
#excuse you from the conditions of this License.  If you cannot
#distribute so as to satisfy simultaneously your obligations under this
#License and any other pertinent obligations, then as a consequence you
#may not distribute the Program at all.	For example, if a patent
#license would not permit royalty-free redistribution of the Program by
#all those who receive copies directly or indirectly through you, then
#the only way you could satisfy both it and this License would be to
#refrain entirely from distribution of the Program.
#
#If any portion of this section is held invalid or unenforceable under
#any particular circumstance, the balance of the section is intended to
#apply and the section as a whole is intended to apply in other
#circumstances.
#
#It is not the purpose of this section to induce you to infringe any
#patents or other property right claims or to contest validity of any
#such claims; this section has the sole purpose of protecting the
#integrity of the free software distribution system, which is
#implemented by public license practices.  Many people have made
#generous contributions to the wide range of software distributed
#through that system in reliance on consistent application of that
#system; it is up to the author/donor to decide if he or she is willing
#to distribute software through any other system and a licensee cannot
#impose that choice.
#
#This section is intended to make thoroughly clear what is believed to
#be a consequence of the rest of this License.
#^L
#  8. If the distribution and/or use of the Program is restricted in
#certain countries either by patents or by copyrighted interfaces, the
#original copyright holder who places the Program under this License
#may add an explicit geographical distribution limitation excluding
#those countries, so that distribution is permitted only in or among
#countries not thus excluded.  In such case, this License incorporates
#the limitation as if written in the body of this License.
#
#  9. The Free Software Foundation may publish revised and/or new versions
#of the General Public License from time to time.  Such new versions will
#be similar in spirit to the present version, but may differ in detail to
#address new problems or concerns.
#
#Each version is given a distinguishing version number.	If the Program
#specifies a version number of this License which applies to it and "any
#later version", you have the option of following the terms and conditions
#either of that version or of any later version published by the Free
#Software Foundation.  If the Program does not specify a version number of
#this License, you may choose any version ever published by the Free Software
#Foundation.
#
#  10. If you wish to incorporate parts of the Program into other free
#programs whose distribution conditions are different, write to the author
#to ask for permission.	For software which is copyrighted by the Free
#Software Foundation, write to the Free Software Foundation; we sometimes
#make exceptions for this.  Our decision will be guided by the two goals
#of preserving the free status of all derivatives of our free software and
#of promoting the sharing and reuse of software generally.
#
#			    NO WARRANTY
#
#  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
#FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
#OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
#PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
#OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
#TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
#PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
#REPAIR OR CORRECTION.
#
#  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
#WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
#REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
#INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
#OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
#TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
#YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
#PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
#POSSIBILITY OF SUCH DAMAGES.
#
#		     END OF TERMS AND CONDITIONS
#^L
#	    How to Apply These Terms to Your New Programs
#
#  If you develop a new program, and you want it to be of the greatest
#possible use to the public, the best way to achieve this is to make it
#free software which everyone can redistribute and change under these terms.
#
#  To do so, attach the following notices to the program.  It is safest
#to attach them to the start of each source file to most effectively
#convey the exclusion of warranty; and each file should have at least
#the "copyright" line and a pointer to where the full notice is found.
#
#    <one line to give the program's name and a brief idea of what it does.>
#    Copyright (C) 19yy	<name of author>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#Also add information on how to contact you by electronic and paper mail.
#
#If the program is interactive, make it output a short notice like this
#when it starts in an interactive mode:
#
#    Gnomovision version 69, Copyright (C) 19yy name of author
#    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
#    This is free software, and you are welcome to redistribute it
#    under certain conditions; type `show c' for details.
#
#The hypothetical commands `show w' and `show c' should show the appropriate
#parts of the General Public License.  Of course, the commands you use may
#be called something other than `show w' and `show c'; they could even be
#mouse-clicks or menu items--whatever suits your program.
#
#You should also get your employer (if you work as a programmer) or your
#school, if any, to sign a "copyright disclaimer" for the program, if
#necessary.  Here is a sample; alter the names:
#
#  Yoyodyne, Inc., hereby disclaims all copyright interest in the program
#  `Gnomovision' (which makes passes at compilers) written by James Hacker.
#
#  <signature of Ty Coon>, 1 April 1989
#  Ty Coon, President of Vice
#
#This General Public License does not permit incorporating your program into
#proprietary programs.  If your program is a subroutine library, you may
#consider it more useful to permit linking proprietary applications with the
#library.  If this is what you want to do, use the GNU Library General
#Public License instead of this License.
#

__version__ = '$Revision: 2.12 $'[11:-2];

import cgi, sys, string, os, re, errno, time, stat, urllib
from cgi import log
from os import path, environ
from socket import gethostbyaddr
from time import localtime, strftime
from cStringIO import StringIO

def emit_header(charset):
    # print html header
    # tell www client to not cache www page
    print "Content-type: text/html;  charset=" + charset
    print "Cache-Control: no-cache"
    print "Pragma: no-cache"
    print "Expires: 0"
    print "Cache-Control: no-store"
    print
    print '<html>'

def get_scriptname():
    # get name of cgi script
    # Return string with script name
    return environ.get('SCRIPT_NAME', '')

def exists(file):
    # Return 1 if file exists and 0 if not exists
    # file is path + filename
    return os.path.exists(file)


def read_lang_file(file,dict):
    # Read language file and return language dictionary
    f = open(file)
    lines = f.readlines()
    for line in lines:
        line = re.sub('^[ ]*#.*\r*\n*',r'',line)
        if line != '':
            var = re.sub('([^ =]*)([ =].*)\n',r'\1',line)
            value = re.sub('([^ =]*)[ =]*(.*)\n',r'\2',line)
            dict[var] = value
    f.close()
    return dict

def number_hash(line):
    # count number of hash at beginning of a line
    count = 0
    for i in range(len(line)):
        if (line[i] == '#') or (line[i] == ' '):
            if line[i] == '#':
                count = count + 1
        else:
            break
    return count

def number_star(line):
    # count stars at beginning of a line
    count = 0
    for i in range(len(line)):
        if (line[i] == '*') or (line[i] == ' '):
            if line[i] == '*':
                count = count + 1
        else:
            break
    return count

def mod_time(pagename):
    # Return time and date for when a page was last modified
    # IN: pagename, wiki page name
    # RETURN: time and date
    pagename = data_dir + pagename
    modtime = localtime(os.stat(pagename)[stat.ST_MTIME])
    return strftime(datetime_fmt, modtime)


def bullet_list(input_rows,line,ul_level,new_ul_level):
    # make a bullet list of *
    # INPUT:
    # input_rows = list of previous lines
    # line = current line
    # ul_level = level of bullet list
    # new_ul_level = new level of bullet list (can be same as ul_level)
    # RETURN:
    # list of previous lines, current line, current level, new level
    if re.search('^[ ]*\*',line):
        new_ul_level = number_star(line)
        if new_ul_level > ul_level:
            for i in range (ul_level,new_ul_level):
                line = re.sub('(.*)',r'<ul><li>\1',line)
        elif new_ul_level < ul_level:
            tmpline = ''
            for i in range (new_ul_level,ul_level):
                tmpline = '</ul>' + tmpline
            line = tmpline + '<li>' + line
        else:
            line = '<li>' + line
        line = re.sub('^([^\*]*)\*+([^\*]*)',r'\1\2',line)
        ul_level = new_ul_level
    else:
        if ul_level != 0:
            tmpline = ''
            for i in range (0,ul_level):
                tmpline = '</ul>' + tmpline
            input_rows.append(tmpline)
        ul_level = 0
    return [input_rows,line, ul_level, new_ul_level]


def numbered_list(input_rows,line,ol_level,new_ol_level):
    #
    # make a numbered list of #
    # INPUT:
    # input_rows = list of previous lines
    # line = current line
    # ol_level = level of numbered list
    # new_ol_level = new level of numbered list (can be same as ol_level)
    # RETURN:
    # list of previous lines, current line, current level, new level    
    if re.search('^[ ]*#',line):
        new_ol_level = number_hash(line)
        if new_ol_level > ol_level:
            for i in range (ol_level,new_ol_level):
                line = re.sub('(.*)',r'<ol><li>\1',line)
        elif new_ol_level < ol_level:
            tmpline = ''
            for i in range (new_ol_level,ol_level):
                tmpline = '</ol>' + tmpline
            line = tmpline + '<li>' + line
        else:
            line = '<li>' + line
        line = re.sub('^([^#]*)#+([^#]*)',r'\1\2',line)
        ol_level = new_ol_level
    else:
        if ol_level != 0:
            tmpline = ''
            for i in range (0,ol_level):
                tmpline = '</ol>' + tmpline
            input_rows.append(tmpline)
        ol_level = 0
    return [input_rows,line,ol_level,new_ol_level]


def images(line):
    # images
    # make a <img src=URL> of a img src=URL
    # line = current line
    if re.search('img src=',line):
        line = re.sub('img src=([^\n\r      ]+)',r'<img src="\1">',line)
    return line


def wiki_links(line,server):
    # convert [] to a wiki link
    # convert [text] to a link to a wiki dokument
    tmpline = ''
    if re.search('\[[^\]]*\]',line):
        # all text to (including) first wiki link
        start_line = re.sub('^([^\[]*\[.*?\]).*$',r'\1',line)
        rest = re.sub('^[^\[]*\[.*?\](.*$)',r'\1',line)
        # all text after first wiki link 
        while start_line != '':
            if re.search('.*\[.*\].*',start_line):
                before_link = re.sub('^([^\[]*)\[.*',r'\1',start_line)
                link = re.sub('^.*\[([^\]]*).*',r'\1',start_line)
                after_link = re.sub('^.*\[.*\](.*)[\r\n]*',r'\1',start_line)
                if re.search('^[Ii][Mm][Gg]:',link):
                    # img, image link to a local picture
                    tmpline = tmpline + before_link + '<img src="http://' + server + image_dir + urllib.quote(re.sub('^[Ii][Mm][Gg]:(.*)$',r'\1',link)) + '">' + after_link
                elif re.search('^[Ii][Mm][Gg]-[Ll][Ii][Nn][Kk]:',link):
                    # imgage link to a picture on the web
                    tmpline = tmpline + before_link + '<img src="http://' + urllib.quote(re.sub('^[Ii][Mm][Gg]-[Ll][Ii][Nn][Kk]:(.*)$',r'\1',link)) + '">' + after_link
                elif re.search('^[Ss][Uu][Ss][Nn][Ii][Nn][Gg]:',link):
                    # link to susning.nu
                    tmpline = tmpline + before_link + '<a href="http://www.susning.nu/' + urllib.quote(re.sub('^[Ss][Uu][Ss][Nn][Ii][Nn][Gg]:(.+)[ ]*$',r'\1',link)) + '">' + re.sub('^[Ss][Uu][Ss][Nn][Ii][Nn][Gg]:(.+)[ ]*$',r'\1',link) + '</a>' + after_link
                elif re.search('^[Rr][Ff][Cc]:',link):
                    # link to rfc page at ftp.sunet.se
                    tmpline = tmpline + before_link + '<a href="http://ftp.sunet.se/pub/Internet-documents/rfc/rfc' + urllib.quote(re.sub('^[Rr][Ff][Cc]:([^ ]+)[ ]*$',r'\1',link)) + '.txt">RFC ' + urllib.quote(re.sub('^[Rr][Ff][Cc]:([^ ]+)[ ]*$',r'\1',link)) + '</a>' + after_link
                elif re.search('^[Ll][Ii][Nn][Uu][Xx]-[Mm][Aa][Nn]:',link):
                    # link to linux man page
                    man_page = urllib.quote(re.sub('^[Ll][Ii][Nn][Uu][Xx]-[Mm][Aa][Nn]:[ ]*(.*)\(.*\)[ ]*$',r'\1',link))
                    man_chapter = urllib.quote(re.sub('^[Ll][Ii][Nn][Uu][Xx]-[Mm][Aa][Nn]:[ ]*.*\((.*)\)[ ]*$',r'\1',link))
                    tmpline = tmpline + before_link + '<a href="http://www.die.net/doc/linux/man/man' + man_chapter + '/' + man_page + '.' + man_chapter + '.html">' + man_page + ' (' + man_chapter + ')' + '</a>' + after_link
                elif re.search('^[Ff][Rr][Oo][Mm]:',link):
                    # from page link. From page is used on menu page
                    tmpline = tmpline + before_link + '<a href="' + scriptname + '?page=' + urllib.quote(re.sub('^[Ff][Rr][Oo][Mm]:[ ]*(.+)[ ]*.*$',r'\1',link))  + '">' + re.sub('^[Ff][Rr][Oo][Mm]:[ ]*(.+)[ ]*.*$',r'\1',link) + '</a>' + after_link
                elif re.search('[Mm][Ee][Nn][Uu]:',link):
                    # menu link
                    tmpline = tmpline + before_link + '<a href="' + scriptname + '?page=' + urllib.quote(re.sub('^[Mm][Ee][Nn][Uu]:[ ]*(.+)[ ]*.*$',r'\1',link))  + '">' + re.sub('^[Mm][Ee][Nn][Uu]:[ ]*(.+)[ ]*.*$',r'\1',link) + '</a>' + after_link
                else:
                    # Wiki page link
                    if nonexist_qm == '0' or exists(data_dir + link):
                        tmpline = tmpline + before_link + '<a href="' + scriptname + '?page=' + urllib.quote(link) + '">' + link + '</a>' + after_link
                    else:
                        tmpline = tmpline + before_link + link + '<a href="' + scriptname + '?page=' + urllib.quote(link) + '">?</a>' + after_link
                start_line  = re.sub('^([^\[]*\[.*?\]).*$',r'\1',rest)
                rest = re.sub('^[^\[]*\[.*?\](.*$)',r'\1',rest)
            else:
                tmpline = tmpline + rest
                start_line = ''
        return tmpline
    else:
        return line

def wiki_header(line):
    # Make a header of a line
    # !!!!! = <h5>
    # !!!!  = <h4>
    # !!!   = <h3>
    # !!    = <h2>
    # !     = <h1>
    # INPUT: line, current line
    # RETURN: line, modified current line
    if re.search('^[ ]*!+[^!]+!+[ ]*[\r\n]*$',line):
        header_text = re.sub('^[ ]*!+([^!]+)!+[ ]*[\r\n]*$',r'\1',line)
        if re.search('^[ ]*!!!!![^!]+!+',line):
            line = '<h5>' + header_text + '</h5>'
        if re.search('^[ ]*!!!![^!]+!+',line):
            line = '<h4>' + header_text + '</h4>'
        elif re.search('^[ ]*!!![^!]+!+',line):
            line = '<h3>' + header_text + '</h3>'
        elif re.search('^[ ]*!![^!]+!+',line):
            line = '<h2>' + header_text + '</h2>'
        elif re.search('^[ ]*![^!]+!+',line):
            line = '<h1>' + header_text + '</h1>'
        return line
    else:
        return line

def wiki_bold_italic(line):
    #
    # convert ''text'' to italic text
    # '''text''' to bold
    # and '''''text''''' to bold italic
    # line = curent line
    # RETURN: modified current line
    if re.search("'",line):
        start_line = re.sub("^([^']*)'.*",r"\1",line)
        rest = re.sub("[^']*('.*)",r"\1",line)
        tmpline = ""
        if start_line != "":
            tmpline = tmpline + start_line
        while rest != "":
            # make a loop over line
            if re.search("'",rest):
                if re.search("^'''''[^']*'''''",rest):
                    # convert '''''text''''' to bold italic text
                    # <b><em>text</em></b>
                    text = re.sub("^'''''([^']*)'''''.*$",r"<b><em>\1</em></b>",rest)
                    tmpline = tmpline + text
                elif re.search("^'''[^']*'''",rest):
                    # convert '''text''' to bold text
                    text = re.sub("^'''([^']*)'''.*$",r"<b>\1</b>",rest)
                    tmpline = tmpline + text
                elif re.search("^''[^']*''",rest):
                    # convert ''text'' to italic text
                    text = re.sub("^''([^']*)''.*$",r"<em>\1</em>",rest)
                    tmpline = tmpline + text
                start_line = re.sub("^[']*[^']*[']*([^']*)'.*$",r"\1",rest)
                rest = re.sub("^'+[^']*'+(.*)$",r"\1",rest)
                rest = re.sub("[^']*('.*)",r"\1",rest)
                if start_line != "":
                    tmpline = tmpline + start_line
                else:
                    tmpline = tmpline + rest
                    rest = ""
            else:
                tmpline = tmpline + start_line
                start_line = ""
                rest = ""
        return tmpline
    else:
        return line

def quot(line):
    # convert " to &quot;  & to &amp;  < to &lt;  > to &gt;
    line = string.replace(line,'&','&amp;')
    line = string.replace(line,'"','&quot;')
    line = string.replace(line,'<','&lt;')
    line = string.replace(line,'>','&gt;')
    return line



def URL_and_quot(line):
    #
    # converts URL to a <a href=URL> link
    # INPUT: current line
    # RETURN: modified current line
    tmpline = ''
    # Remove return and line feed
    line = re.sub('^(.*?)[\n\r]*$',r'\1',line)
    if re.search('http://',line) or re.search('ftp://',line) or re.search('gopher://',line) or re.search('file:/',line):
        before_link = re.sub('^(.*?)\w+://[^ "\(\)\n\r]+[ \)"]*.*',r'\1',line)
        link = re.sub('^.*?[ \("]*(\w+://[^ \(\)"]+)[ \)"]*.*',r'<a href="\1">\1</a>',line)
        after_link = re.sub('^.*?\w+://[^ \(\)"]+([ \)"]*.*)',r'\1',line)
        if after_link == '':
            tmpline = before_link + link + after_link
        while after_link != '':
            if re.search('[ \("]*\w+://.+?[ \)"]*',after_link):
                tmpline = tmpline + quot(before_link)
                tmpline = tmpline + link
                before_link = re.sub('^(.*?)\w+://[^ \(\)"\n\r]+[ \)"]*.*',r'\1',after_link)
                link = re.sub('^.*?[ \("](\w+://[^ \(\)"]+)[ \)"]*.*',r'<a href="\1">\1</a>',after_link)
                after_link = re.sub('^.*?\w+://[^ \(\)"]+([ \)"]*.*)',r'\1',after_link)
                if before_link == link:
                    tmpline = before_link
                    break
                if re.search('^[\n\r]*$',after_link):
                    tmpline = tmpline + quot(before_link)
                    tmpline = tmpline + link
            else:
                if after_link != '':
                    tmpline = tmpline + quot(before_link)
                    tmpline = tmpline + link
                    tmpline = tmpline + quot(after_link)
                after_link = ''
        return tmpline
    else:
        return quot(line)


def wiki_table(line,table):
    # make a table with border = 0 of
    #                   || cell || cell ||
    #                   || cell || cell ||
    #                   || cell || cell ||
    # and table with border = 1 of
    #                   ||| cell ||| cell |||
    #                   ||| cell ||| cell |||
    #                   ||| cell ||| cell |||
    #
    # input: line=current line
    #        table, boolean, 0 if previous line was not a table row
    #                        1 if previous line was a table row
    # return list with line and table, [line,table]
    #
    start_tr = 0
    if re.search('^[^\|]*\|\|\|?[^\|]*\|\|\|?.*$',line):
        before_table = re.sub('^([^\|]*)\|.*',r'\1',line)
        rest = re.sub("[^\|]*(\|.*)",r"\1",line)
        tmpline = ""
        if before_table != "":
            tmpline = tmpline + before_table
        while rest != "":
            # make a loop over line
            if re.search("\|\|\|?[^\|]*\|\|\|?",rest):
                if re.search('^\|\|\|[^\|]*\|\|\|?',rest) and not table:
                    tmpline = tmpline + '<table border=1><tr>'
                    table = 1
                    start_tr = 1
                elif re.search('^\|\|[^\|]*\|\|',rest) and not table:
                    tmpline = tmpline + '<table border=0><tr>'
                    table = 1
                    start_tr = 1
                else:
                    if not start_tr:
                        tmpline = '<tr>' + tmpline
                        start_tr = 1
                        table = 1
                text = re.sub('^\|\|\|?([^\|]*)\|\|\|?.*$',r'<td>\1</td>',rest)
                tmpline = tmpline + text
                before_table = re.sub('^([^\|]*)\|.*$',r"\1",rest)
                rest = re.sub("^\|+[^\|]*(\|+.*)$",r"\1",rest)
                rest = re.sub("[^\|]*(\|.*)",r"\1",rest)
                if before_table != "":
                    tmpline = tmpline + before_table
            else:
                rest = re.sub('\|*(.*)',r'\1',rest)
                tmpline = tmpline + before_table + '</tr>' + rest
                before_table = ""
                rest = ""
    else:
        if table:
            tmpline = '</table>' + line
        else:
            tmpline = line
        table = 0
    return [tmpline,table]


def wiki_to_html(input,server):
    # Turn wiki markup into HTML
    #
    # input is a list of text lines
    # Function returns a modified input list
    ul = ''    
    ul_level = 0
    new_ul_level = 0
    ol_level = 0
    new_ol_level = 0
    pre = 0
    table = 0
    tmpinput = []
    for line in input:

        # Delete new line at end of line
        line = re.sub('(.*)[\n\r]*',r'\1',line)

        # URL_and_quot
        if pre != 1 and not re.search('{{{',line):
            if re.search('^[ ]*img src=.*$',line):
                line = images(line)
            else:
                line = URL_and_quot(line)

        # everything between {{{ and }}} as it is.
        # {{{
        # text
        # }}}
        if re.search('{{{[^{}]*}}}',line):
            start_input = [re.sub('^([^{}]*?){{{.*',r'\1',line)]
            pre_text = re.sub('^[^{}]*?{{{([^{}]+)}}}.*',r'\1',line)
            rest = [re.sub('^[^{}]*?{{{[^{}]*}}}(.*)',r'\1',line)]
            pre_tmp_lines = wiki_to_html(start_input,server)
            for pre_line in pre_tmp_lines:
                tmpinput.append(pre_line)
            tmpinput.append('<pre>' + quot(pre_text) + '</pre>')
            pre_tmp_lines = wiki_to_html(rest,server)
            for pre_line in pre_tmp_lines:
                tmpinput.append(pre_line)
            #line = re.sub('^[\n\r   ]*$',r'<p>',line)
            line = ''
            
        elif re.search('{{{',line):
            pre = 1
            start_input = [re.sub('^([^{}]*?){{{.*',r'\1',line)]
            rest = re.sub('.*?{{{(.*)',r'\1',line)
            if start_input != '':
                pre_tmp_lines = wiki_to_html(start_input,server)
                for pre_line in pre_tmp_lines:
                    tmpinput.append(pre_line)
            tmpinput.append('<pre>' + rest)
            line = ''
        elif pre == 1:
            if re.search('}}}',line):
                start_line = re.sub('(.*?)}}}.*',r'\1',line)
                rest = [re.sub('.*?}}}(.*)',r'\1',line)]
                tmpinput.append(start_line)
                pre_tmp_lines = wiki_to_html(rest,server)
                for pre_line in pre_tmp_lines:
                    tmpinput.append('</pre>' + pre_line)
                pre = 0
            else:
                tmpinput.append(quot(line))
            line = ''
        else:
            # convert empty lines to <p>
            line = re.sub('^[\n\r 	]*$',r'<p>',line)
            # bullet list
            tmplist = bullet_list(tmpinput,line,ul_level,new_ul_level)
            tmpinput = tmplist[0]
            line = tmplist[1]
            ul_level = tmplist[2]
            new_ul_level = tmplist[3]
            
            # numbered list
            tmplist = numbered_list(tmpinput,line,ol_level,new_ol_level)
            tmpinput = tmplist[0]            
            line = tmplist[1]
            ol_level = tmplist[2]
            new_ol_level = tmplist[3]
            
            # convert two [[ to ^G^G
            line = re.sub('\[\[',r'',line)

            # convert [text] to a link to a wiki dokument
            #line = re.sub('\[([^][]*)\]',r'<a href="kpwiki.cgi?page=\1">\1</a>',line)

            line = wiki_links(line,server)
            
            # convert ^G^G to a single [
            line = re.sub('',r'[',line)
            
            # wiki header, !, !!, !!!, !!!!
            line = wiki_header(line)
            
            # wiki bold, italic
            line = wiki_bold_italic(line)
            
            # 5 - on a single line makes a <hr>
            line = re.sub('^[\n\r 	]*-----[\n\r 	-]*$',r'<hr>',line)
            # three %%% at end of line makes a line break
            line = re.sub('^(.*)%%%[\n\r 	%]*$',r'\1<br>',line)

            # table
            tmplist = wiki_table(line,table)
            line = tmplist[0]
            table = tmplist[1]
            
        if line != '':
            tmpinput.append(line)
    if table != 0:
        tmpinput.append('</table>')
        table = 0
    if ul_level != 0:
	# if bullet list at end of page
        line = ''
        for i in range (0,ul_level):
            line = '</ul>'
            tmpinput.append(line)
        ul_level = 0
    if ol_level != 0:
	# if numbered list at end of page
        line = ''
        for i in range (0,ol_level):
            line = '</ol>'
            tmpinput.append(line)
        ol_level = 0
    return tmpinput

def print_short_guide(lang):
    # Print a short guide to KpWiki markup
    #
    lang_shortguide = lang_dir + lang + '.shortguide'
    if exists(lang_shortguide):
        f = open(lang_shortguide)
        lines = f.readlines()
        for line in lines:
            print line + '<br>'
    elif lang == '' or lang == 'en' or lang == 'en_US':
        print 'Write: <br>'
        print '!text! for &lt;h1&gt; header<br>'
        print '!!text!! for &lt;h2&gt; header<br>'
        print '!!!text!!! for &lt;h3&gt; header<br>'
        print '!!!!text!!!! for &lt;h4&gt; header<br>'
        print '[link] for a link to another page<br>'
        print "''text'' for <em>italic</em> text<br>"
        print "'''text''' for <b>bold</b> text<br>"
        print "'''''text''''' for bold italic <b><em>text</em></b><br>"
        print "----- on a single line makes a horizontal line<br>"
        print "%%% at the end of a line makes a line break<br>"
        print "* makes a bullet list.<br>"
        print 'Write two [[ for a ['

def sort_list(list):
    # bubblesort. Not most effective but it works
    # list with name and number of times viewed
    l = list[:]
    for i in range(len(l)):
        for j in range(i+1,len(l)):
            if l[i][1] > l[j][1]:
                l[i],l[j] = l[j],l[i]
    return l

def sort_ignorecase(list):
    # bubblesort
    l = list[:]                       # create a slice-copy of the list
    for i in range(len(l)):           # for every element [i] in the list
        for j in range(i+1,len(l)):   # examine every element [j] after it
            if string.lower(l[i]) > string.lower(l[j]): # and if they are "out of order"
                l[i],l[j]=l[j],l[i]   # swap them
    return l


def insert_links(input, wiki_page_content, wikipagename):
    # input contains menu page
    # wiki_page_content contains visited wiki page content
    # wikipagename visited wiki page name
    #
    # loop
    #    menu page lines
    #    if from link
    #       read wiki page
    #       insert menu links from page
    #
    tmpinput = []
    rest = ''
    for line in input:
        if re.search('\[[^\]]*\]',line):
            start_line = re.sub('^([^\]]*\]).*[\n\r]*$',r'\1',line)
            rest = re.sub('^[^\]]*\](.*$)[\n\r]*',r'\1',line)
            while start_line != '':
                if re.search('\[[^\]]*\]',start_line):
                    before_link = re.sub('^([^\[\n\r]*)\[.*',r'\1',start_line)
                    link = re.sub('^.*\[([^\]]*)[ 	\n\r]*.*',r'\1',start_line)
                    after_link = re.sub('^.*\[.*\]([^\n\r]*)[\n\r]*',r'\1',start_line)
                    if re.search('^[Ff][Rr][Oo][Mm]:',link):
                        tmpinput.append(before_link)
                        frompagename = re.sub('^[Ff][Rr][Oo][Mm]:[ ]*([^\n\r]+)[ 	\n\r]*$',r'\1',link)
                        frompage_content = []
                        if frompagename != wikipagename:
                            frompage_file = data_dir + frompagename
                            if exists(frompage_file):
                                from_page = open(data_dir + frompagename,'r')
                                frompage_content = from_page.readlines()
                            else:
                                frompage_content.append('[' + link + ']')
                        else:
                            frompage_content = wiki_page_content
                        for frompage_line in frompage_content:
                            if  re.search('\[[Mm][Ee][Nn][Uu]:',frompage_line):
                                tmpinput.append('* [' + re.sub('^.*\[[Mm][Ee][Nn][Uu]:[ ]*([^\n\r\]]+)[ 	]*.*$',r'\1',frompage_line) + ']')
                            else:
                                pass
                        tmpinput.append(after_link)
                    else:
                        tmpinput.append(start_line)
                    start_line  = re.sub('^([^\]]*\]).*$',r'\1',rest)
                    rest = re.sub('^[^\]]*\](.*$)',r'\1',rest)
                else:
                    if not re.search('^[\n\r]*$',start_line):
                        tmpinput.append(start_line)
                    start_line = ''
        else:
            tmpinput.append(line)
            start_line = ''
    for line in tmpinput:
        if line == '\n':
            tmpinput.remove(line)
        if line == '\r':
            tmpinput.remove(line)
        if line == '':
            tmpinput.remove(line)
    return tmpinput


def show_left_menu(pagename, wiki_page_content, wikipagename):
    # print a menu on the left side of the page
    server = environ.get('HTTP_HOST')
    left_menu_file = data_dir + pagename
    if exists(left_menu_file):
        infile = open(left_menu_file,'r')
        input = infile.readlines()
        if pagename == 'left_menu':
            input = insert_links(input, wiki_page_content, wikipagename)
        rcs_revision = ''
        if re.search('\$Revision:',input[0]):
            rcs_revision = re.sub('\$Revision:[ ]*([^\$]*)\$',r'\1',input[0])
            del input[0]
        input = wiki_to_html(input,server)
        for i in range(len(input)):
            print input[i]
        print '<p>'
    else:
        print '<p>'


def print_menu_table(form, wikipagename, wiki_page_content, lang,lang_dict):
    # print table and
    # left cell (menu)
    if exists(data_dir+'left_menu'):
        print '<table border=0 width=100% height=100%><tr><td valign=top BGCOLOR=#bbffff width=' + menu_width + '%>'
        print '<center>' + menu_logo_string + '</center>'
    pagename = 'left_menu'
    show_left_menu(pagename,wiki_page_content, wikipagename)
    if exists(data_dir + pagename):
        print '</td><td rowspan=2 valign=top>'

def print_end_menu_table():
    # print end of table
    pagename = 'left_bottom_menu'
    print '</td></tr><tr><td valign=bottom BGCOLOR=#bbffff>'
    if exists(data_dir + pagename):
        show_left_menu(pagename,[],"")
    else:
        print '<p></p>'
    print '</td></tr></table>'



def edit_default_lang_dict():
    # default language dictionary for edit page
    dict = {}
    dict['edit'] = 'Edit'
    dict['save'] = 'Save'
    dict['cancel'] = 'Cancel'
    dict['savetext'] = 'Save page'
    dict['more_help'] = 'More Help'
    dict['edit_locked'] = 'File locked by another user.'
    dict['edit_locked_override_text'] = 'If that not the case remove lockfile'
    dict['edit_locked_override'] = 'Remove lockfile'
    dict['new_page'] = 'Describe "%s" here.'
    dict['read_only'] = 'Read only page'
    return dict

def read_only(pagename):
    # is this wiki page a read only page?
    if site_lock == '1':
        return(1) # Yes, entire site is read only
    else:
        if exists(kpwiki_readonly_pages):
            f = open(kpwiki_readonly_pages)
            lines=f.readlines()
            for line in lines:
                line = re.sub('^[ ]*#.*\r*\n*',r'',line)
                if re.search('^' + pagename + '$',line):
                    return(1)
            return(0)
        else:
            return(0)

def print_readonly_message(form,pagename, lang_dict):
    thispage = 'edit_page'
    print '<head><title>KpWiki: ' + lang_dict['edit'] + ': ' + pagename + '</title>'
    if css_url:
        print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
    print '</head><body>'
    if left_menu == 'on':
        wiki_page_content = []
        thispage = 'edit_page'
        print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
    print '<h1>'
    if left_menu != 'on':
            print logo_string
            print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    print lang_dict['edit'] + ': ' + pagename + '</h1>'
    print '<hr>'
    print lang_dict['read_only']
    print '<p>'
    print '<p><hr></form>'
    if left_menu == 'on':
        print_end_menu_table()
    print '</body></html>'    
    


def remove_readonly_pages(pagelist):
    newlist = []
    for page in pagelist:
        if not read_only(page):
            newlist.append(page)
    return newlist

def edit_page(form,pagename,lang,lang_dict_main):
    # Edit page
    #
    lang_dict = edit_default_lang_dict()   # default language dictionary for edit
    lang_edit = lang_dir + lang + '.edit'
    if exists(lang_edit):
        lang_dict = read_lang_file(lang_edit, lang_dict)
    if read_only(pagename):
        print_readonly_message(form,pagename,lang_dict)
        return(0)
    lockfile = log_dir + pagename + '.LCK'
    tmplockfile = path.join(log_dir, ('#tmp-' + pagename))
    tmpfile = open(tmplockfile,'w')
    portalocker.lock(tmpfile, portalocker.LOCK_EX)
    if form.has_key('old_page'):
        old_page = form['old_page'].value
        if re.search('\.\.',old_page):
            # Do not allow .. in page names
            old_page = urllib.quote(scriptname) + '?page=' + urllib.quote(frontpage)
    elif os.environ.has_key('HTTP_REFERER'):
        http_referer = os.environ['HTTP_REFERER']
        if re.search('\.\.',http_referer):
            # Do not allow ..  in page names
            old_page = urllib.quote(scriptname) + '?page=' + urllib.quote(frontpage)
        else:
            old_page = http_referer
    else:
        old_page = urllib.quote(scriptname) + '?page=' + urllib.quote(frontpage)
    if not exists(lockfile):
        file = open(lockfile,'w')
        portalocker.lock(file, portalocker.LOCK_EX)
        file.close()
        tmpfile.close()
        os.remove(tmplockfile)
        print '<head><title>KpWiki: ' + lang_dict['edit'] + ': ' + pagename + '</title>'
        if css_url:
            print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
        print '</head><body>'
        if left_menu == 'on':
            wiki_page_content = []
            thispage = 'edit_page'
            print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
        print '<h1>'
        if left_menu != 'on':
            print logo_string
            print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
        print lang_dict['edit'] + ': ' + pagename + '</h1>'
        print '<form method="post" action="%s">' % (scriptname)
        print '<input type=hidden name="savepage" value="%s">' % urllib.quote(pagename)
        print '<input type=submit value="' + lang_dict['save'] + '">'
        print lang_dict['savetext'] + ' %s &nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp; ' % pagename
        #print '<input type=submit value="' + lang_dict['cancel'] + '" name="edit_lock_override">'
        print '<input type=hidden name="edit_lock_file" value="' + lockfile + '">'
        print '<input type=hidden name="page" value="' + urllib.quote(pagename) + '">'
        print '<hr>'
        raw_body = ''
        rcs_revision = ''
        page = path.join(data_dir, pagename)
        if exists(page):
            file = open(page,'r')
            input = file.readlines()
            if re.search('\$Revision:',input[0]):
                rcs_revision = re.sub('(\$Revision:[ ]*[^\$]*\$)',r'\1',input[0])
                del input[0]
            else:
                rcs_revision = '$Revision: 1.0 $'
            for i in range(len(input)):
                raw_body = raw_body + input[i]
        else:
            raw_body = lang_dict['new_page'] % pagename
            print '<input type=hidden name="edit_new_page" value="' + urllib.quote(pagename) + '">'
        print '<input type=hidden name="old_page" value="' + old_page + '">'
        print '<input type=hidden name="rcs_revision" value="' + urllib.quote(rcs_revision) + '">'
        print """<textarea wrap="virtual" name="savetext" rows="26" cols="90">%s</textarea>""" % raw_body
        print '<hr><input type=submit value="' + lang_dict['save'] + '">'
        print lang_dict['savetext'] + ' %s &nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp; &nbsp;&nbsp; &nbsp;&nbsp; &nbsp; &nbsp;' % pagename
        print '<input type=submit value="' + lang_dict['cancel'] + '" name="edit_lock_override">'
        print '<p>'
        print_short_guide(lang)
        print '<br><a href="' + scriptname + '?page=edit help">' + lang_dict['more_help'] + '</a><br>'
        print '</form>'
        if left_menu == 'on':
            print_end_menu_table()
        print '</body></html>'
    else:
        tmpfile.close()
        if exists(tmplockfile):
            os.remove(tmplockfile)
        print '<head><title>KpWiki: ' + lang_dict['edit'] + ': ' + pagename + '</title>'
        if css_url:
            print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
        print '</head><body>'
        if left_menu == 'on':
            wiki_page_content = []
            thispage = 'edit_page'
            print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
        print '<h1>'
        if left_menu != 'on':
            print logo_string
            print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
        print lang_dict['edit'] + ': ' + pagename + '</h1>'
        print '<form method="post" action="%s">' % (scriptname)
        print lang_dict['edit_locked'] + '<p>'
        print lang_dict['edit_locked_override_text'] + '<p>'
        print '<input type=submit value="' + lang_dict['edit_locked_override'] + '" name="edit_lock_override">'
        print '<input type=hidden name="edit_lock_file" value="' + lockfile + '">'
        print '<input type=hidden name="page" value="' + urllib.quote(pagename) + '">'
        print '<p><hr></form>'
        if left_menu == 'on':
            print_end_menu_table()
        print '</body></html>'


def show_most_viewed(form,lang,lang_dict):
    # Shows a list over the most viewed pages
    #
    # takes form, lang string and lang_dict dictionary
    #
    print '<h2>' + lang_dict['viewed'] + '</h2>'
    count_dir = log_dir + 'count/'
    list = []
    files = os.listdir(count_dir)
    for page in files:
        pagefile = count_dir + page
        file = open(pagefile, 'r')
        portalocker.lock(file, portalocker.LOCK_EX)
        count = string.atoi(file.read())
        l = [page,count]
        list.append(l)
        file.close()
    list = sort_list(list)
    list.reverse()
    print '<table border=1><tr><th>' + lang_dict['pagename'] + '</th><th>' + lang_dict['no_view'] + '</th></tr>'
    for page in list:
        print '<tr><td>'
        print '<a href="' + scriptname + '?page=' + urllib.quote(page[0]) + '">' + page[0] + '</a> &nbsp;'
        print '</td><td>'
        print page[1]
        print '</td></tr>'
    print '</table>'
    if len(list) == 0:
        print lang_dict['view_error']
    print '<hr>'

def show_recent_changes(lang,lang_dict):
    # Reads edit log and shows a list over the most recent changes
    #
    print '<h2>' + lang_dict['recentchanges'] + '</h2>'
    input = []
    count = 0
    if exists(editlog_name):
        file = open(editlog_name,'r')
        input = file.readlines()
        file.close()
        input.reverse()
        print '<table border=1><tr><th>' + lang_dict['pagename'] + '</th><th>' + lang_dict['hostname'] + '</th><th>' + lang_dict['time'] + '</th></tr>'
        for line in input:
            count = count + 1
            # Delete new line at end of line
            line = re.sub('(.*)\n',r'\1',line)
            if not re.search('^ *ERROR',line):
                tmp_page, tmp_host, tmp_date = string.split(line,'\t')
                tmp_page = re.sub('^[ 	]*(.*)',r'\1',tmp_page)
                time_tuple = localtime(float(tmp_date))
                day = tuple(time_tuple[0:3])
                print '<tr><td><a href="' + scriptname + '?page=' + urllib.quote(tmp_page) + '">' + tmp_page + '</a>&nbsp;</td>'
                print '<td>' + tmp_host + '&nbsp;</td>'
                print '<td>'
                print strftime(datetime_fmt, time_tuple)
                print ' &nbsp;</td></tr>'
            if count > 20:
                break
        print '</table>'
    else:
        print lang_dict['recentchanges_error']
    print '<hr>'

def search_default_lang_dict():
    # Default texts for search page
    #
    dict={}
    dict['search'] = 'Search'
    dict['searchtext'] = 'Search for'
    dict['searchhelp'] = 'OBS! KpWiki lowercase all text on all pages before it matches what you wrote.<br>You can use regular expressions, such as .*seriali[sz]e'
    dict['nothing_found'] = 'Did not match any documents. No pages were found containing'
    dict['search_sum1'] = 'hits out of'
    dict['search_sum2'] = 'pages searched. '
    dict['viewed'] = 'Show most viewed pages'
    dict['pagename'] = 'Pagename'
    dict['no_view'] = 'Viewed number of times'
    dict['view_error'] = 'No pages viewed'
    dict['recentchanges'] = 'Show most recent changes'
    dict['recentchanges_error'] = 'No pages edited'
    dict['hostname'] = 'Changed by<br>user from host'
    dict['time'] = 'Time and date'
    dict['checkout1'] = 'Checkout revision'
    dict['checkout2'] = 'of page'
    dict['checkout3'] = 'Press the link for the revision you want to check out.'
    dict['checkout4'] = 'Revision'
    dict['checkout5'] = 'of page checked out.'
    return dict

def remove_characters_from_search(line):
    # Remove characters from search result
    line = re.sub('^(.*)%%%[ \n\r]*',r'\1',line)
    return line

def search(form,lang,lang_dict_main):
    # Print a search page
    # If user searches, the funcion read all pages and do a regexp search
    # Before it does a regexp search it lowercases the page
    # Search also does a regexp search on page names
    #  
    lang_dict = search_default_lang_dict()   # default language dictionary for search
    lang_search = lang_dir + lang + '.search'
    if exists(lang_search):
        lang_dict = read_lang_file(lang_search, lang_dict)
    text = ''
    search_hit = 0
    no_pages = 0   # number of pages
    print '<head><title>KpWiki: ' + lang_dict['search'] + '</title>'
    if css_url:
        print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
    print '</head><body>'
    if left_menu == 'on':
        wiki_page_content = []
        thispage = 'search'
        print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
    print '<h1>'
    if left_menu != 'on':
        print logo_string
        print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    print '<a href="' + scriptname + '?page=search">' + lang_dict['search'] + '</a></h1>'
    print 'KpWiki ' +  lang_dict_main['version'] + ' ' + __version__ + ' ' + copyright + '<hr>'
    if form.has_key('searchtext'):
        files = os.listdir(data_dir)
        files.remove('RCS')
        for pagename in files:
            no_pages = no_pages + 1
            hit = 0
            page_first_lines = []  # the first lines of the wiki page
            no_lines = 0  # number of read lines
            max_no_lines = string.atoi(show_no_lines_on_hit) + 1  # maximum number of lines to read from wiki page and show on a search hit
            tempname = string.lower(pagename)
            if re.search(string.lower(urllib.unquote(form['searchtext'].value)),tempname):
                hit = 1
                search_hit = search_hit + 1
                f = open(data_dir + pagename,'r')
                input = f.readlines()
                if len(input) < max_no_lines:
                    max_no_lines = len(input) - 1
                for no_lines in range(1,max_no_lines):
                    page_first_lines.append(input[no_lines])
                f.close()
            else:
                f = open(data_dir + pagename,'r')
                input = f.readlines()
                if len(input) < max_no_lines:
                    max_no_lines = len(input) - 1
                for no_lines in range(1,max_no_lines):
                    page_first_lines.append(input[no_lines])
                if hit == 0:
                    for line in input:
                        line = string.lower(line)
                        if re.search(string.lower(urllib.unquote(form['searchtext'].value)),line):
                            hit = 1
                            search_hit = search_hit + 1
                            break
                f.close()
            if hit == 1:
                #print '<a href="' + scriptname + '?page=' + urllib.quote(pagename) + '">' + pagename + '</a><br>'
                print '<br><b><a href="' + scriptname + '?page=' + urllib.quote(pagename) + '">' + pagename + '</a></b><br><small>'
                for no_lines in range(0,len(page_first_lines)):
                    print quot(remove_characters_from_search(page_first_lines[no_lines])),
                    print '<br>'
                print '</small>'
        if search_hit == 0:
            print lang_dict['nothing_found'] + ': <b>' + urllib.unquote(form['searchtext'].value) + '</b><p>'
        no_pages = no_pages + 1    # include index page in total nuber of pages
        print '<p>' + repr(search_hit) + ' ' + lang_dict['search_sum1'] + ' ' + repr(no_pages) + ' ' + lang_dict['search_sum2']
        print '<hr>'
    elif form.has_key('viewed'):
        show_most_viewed(form,lang,lang_dict)
    elif form.has_key('recentchanges'):
        show_recent_changes(lang,lang_dict)
    print '<form  method="post" action="%s">' % (scriptname)
    print lang_dict['searchtext'] + ': <input name=searchtext size=30 value="">'
    print '<input type=submit value="' + lang_dict['search'] + '"><p>'
    print lang_dict['searchhelp']
    print '<p><br><br>'
    print '<input type=submit value="' + lang_dict['viewed'] + '" name="viewed"><br>'
    print '<input type=submit value="' + lang_dict['recentchanges'] + '" name="recentchanges"><p>'
    print '<input type=hidden name="search" value="-">'
    print '<hr>'
    print '<input type=submit value="' + lang_dict_main['goto'] + '"><input name=goto size=25 value="">'
    print '<br>'
    print '<a href="' + scriptname + '?index=index">' + lang_dict_main['index'] +  '</a><br>'
    print '<a href="' + scriptname + '?page=' + frontpage +'">' + lang_dict_main['frontpage'] + '</a><br>'
    print '<a href="' + scriptname + '?page=admin">' + lang_dict_main['admin'] + '</a><br>'
    print '</form>'
    if left_menu == 'on':
        print_end_menu_table()
    print '</body></html>'


def makeindex(form,pagename,lang,lang_dict_main):
    # prints an index page
    #
    count = 0
    letter = ''
    new_table = 0
    print '<head><title>KpWiki: ' + pagename + '</title>'
    if css_url:
        print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
    print '</head><body>'
    if left_menu == 'on':
        wiki_page_content = []
        thispage = 'index'
        print_menu_table(form, thispage, wiki_page_content, lang,lang_dict_main)
    print '<h1>'
    if left_menu != 'on':
        print logo_string
        print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    print '<a href="' + scriptname + '?page=' + pagename + '">' + pagename + '</a></h1>'
    print 'KpWiki ' +  lang_dict_main['version'] + ' ' + __version__ + ' ' + copyright + '<hr>'
    files = os.listdir(data_dir)
    files.append('index')
    files.remove('RCS') # Do not include RCS directory
    files = sort_ignorecase(files)
    print '<h2>'
    for page in files:
        firstletter = string.upper(page[0])
        if firstletter != letter:
            letter = string.upper(page[0])
            print firstletter + ' | '
    print '</h2>'
    letter = ''
    for page in files:
        count = count + 1
        firstletter = string.lower(page[0])
        if firstletter != letter:
            if new_table == 1:
                print '</table>'
            letter = string.lower(page[0])
            print '<h2><a NAME="' + firstletter + '">' + firstletter + '</a></h2>'
            print '<table border=1>'
            new_table = 1
        print '<tr><td>' + repr(count) + '</td><td><a href="' + scriptname + '?page=' + urllib.quote(page) + '">' + page + '</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td>'
        print '<td>&nbsp;&nbsp;&nbsp;<a href="' + scriptname + '?info=' + urllib.quote(page) + '">' + lang_dict_main['info_about_page'] + '</a></td></tr>'
    if new_table == 1:
        print '</table>'
    print '<p><hr>'
    print '<form method="post" action="%s">' % (scriptname)
    #print '<a href="' + scriptname + '?index=' + urllib.quote(pagename) + '?wordindex=' + urllib.quote(pagename) + '">' + lang_dict_main['wordindex'] + '</a><br>'
    print '<a href="' + scriptname + '?search=' + urllib.quote(pagename) + '">' + lang_dict_main['search'] + '</a>'
    print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<input type=submit value="' + lang_dict_main['goto'] + '"><input name=goto size=25 value="">'
    print '<br>'
    print '<a href="' + scriptname + '?page=' + urllib.quote(frontpage) +'">' + lang_dict_main['frontpage'] + '</a><br>'
    print '<a href="' + scriptname + '?page=admin">' + lang_dict_main['admin'] + '</a><br>'
    print '</form>'
    if left_menu == 'on':
        print_end_menu_table()
    print '</body></html>'


def info_about_page(form,lang,pagename,lang_dict_main):
    # prints more information about a page (info about page)
    # Prints last modified time and date for page
    # Prints log messages (rcs: rlog page) and other information about the page
    # Prints last lines from edit log about the page
    # Prints page revisions and gives the ability to check out old versions
    #
    rcs_info_available = 1 # rcs information is available, 0 if not available
    lang_dict = search_default_lang_dict()   # default language dictionary for search
    lang_search = lang_dir + lang + '.search'
    if exists(lang_search):
        lang_dict = read_lang_file(lang_search, lang_dict)
    print '<head><title>KpWiki: ' + pagename + '</title>'
    if css_url:
        print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
    print '</head><body>'
    if left_menu == 'on':
        wiki_page_content = []
        thispage = 'infopage'
        print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
    print '<h1>'
    if left_menu != 'on':
        print logo_string
        print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    print '<a href="' + scriptname + '?page=' + urllib.quote(pagename) + '">' + pagename + '</a></h1>'
    print 'KpWiki ' +  lang_dict_main['version'] + ' ' + __version__ + ' ' + copyright + '<hr>'
    rcs_head = ''
    rcs_revision = ''
    page_revisions = []
    page = data_dir + pagename
    if not exists(page):
        print lang_dict_main['info_error']
        print '<p>'
    else:
        print  pagename + ' ' + lang_dict_main['last_mod'] + ' ' + mod_time(pagename) + '<p>'
        pagename = re.sub("'","",pagename)   # remove ' from pagename
        rcs_command = "cd " + data_dir + "; rlog '" + pagename + "'"
        p = os.popen(rcs_command)
        input = p.readlines()
        if len(input) != 0:
            print '<p>'
            print '<pre>'
            for line in input:
                # Delete new line at end of line
                inputline = re.split('\n',line)
                if re.search('head:',inputline[0]):
                    rcs_head = re.sub('head:[ 	]*([0-9\.]*)',r'\1',inputline[0])
                if re.search('revision[ 	]*[0-9\.]+',inputline[0]):
                    rcs_revision = re.sub('revision[ 	]*([0-9\.]*)',r'\1',inputline[0])
                    page_revisions.append(rcs_revision)
                    beginning = re.sub('([^0-9\.]*)[ 	]*[0-9\.]*$',r'\1',inputline[0])
                    print beginning + '<a href="' + scriptname + '?info=' + urllib.quote(pagename) + '&rcs_revision=' + urllib.quote(rcs_revision) + '&rcs_head=' + urllib.quote(rcs_head) + '">' + rcs_revision + '</a>'
                else:
                    print inputline[0]
            print '</pre>'
            print '<p>' + lang_dict_main['rcs_diff_help'] + '<p>'
        else:
            rcs_info_available = 0
            print lang_dict_main['no_rcs_info']
        if form.has_key('rcs_revision') and form.has_key('rcs_head'):
            print '<p>'
            first_release = urllib.unquote(form['rcs_revision'].value)
            second_release = urllib.unquote(form['rcs_head'].value)
            if re.search('[0-9]+\.[0-9]+',first_release) and re.search('[0-9]+\.[0-9]+',second_release):
                pagename = re.sub("'","",pagename)   # remove ' from pagename
                first_release = re.sub("'","",first_release)   # remove ' from first_release
                second_release = re.sub("'","",second_release)   # remove ' from second_release
                rcs_command = "cd " + data_dir + "; rcsdiff -u -r'" + first_release + "' -r'" + second_release + "' '" + pagename + "'"
                p = os.popen(rcs_command)
                input = p.readlines()
                if len(input) != 0:
                    print '<p><hr>'
                    print '<table border=0>'
                    print '<tr><td BGCOLOR=dfdfdf>'
                    print lang_dict_main['rcs_diff_text'] + ': ' + urllib.unquote(form['rcs_revision'].value) + ', ' + urllib.unquote(form['rcs_head'].value) + '<p><br></td></tr>'
                    for line in input:
                        # Delete new line at end of line
                        if re.search('^[ ]*-.*',line):
                            # deleted lines in red
                            print '<tr><td BGCOLOR=ffbbbb><pre>'
                        elif re.search('^[ ]*\+.*',line):
                            # added lines in green
                            print '<tr><td BGCOLOR=bbffbb><pre>'
                        else:
                            # unchanged lines in yellow
                            print '<tr><td BGCOLOR=fcfedd><pre>'
                        inputline = re.split('\n',line)
                        print inputline[0]
                        print '</pre></td></tr>'
                    print '</table>'
                    print '<hr>'
        count = 0
        if exists(editlog_name):
            file = open(editlog_name,'r')
            input = file.readlines()
            file.close()
            input.reverse()
            print '<table border=1><tr><th>' + lang_dict['pagename'] + '</th><th>' + lang_dict['hostname'] + '</th><th>' + lang_dict['time'] + '</th></tr>'
            for line in input:
                if not re.search('^ *ERROR',line):
                    tmp_page, tmp_host, tmp_date = string.split(line,'\t')
                    tmp_page = re.sub('^[ 	]*(.*)',r'\1',tmp_page)
                    time_tuple = localtime(float(tmp_date))
                    day = tuple(time_tuple[0:3])
                    text = re.sub('^ *(.*)',r'\1',tmp_page)
                    tmp_page_encoded = urllib.quote(text)
                    if pagename == text:
                        count = count + 1
                        print '<tr><td><a href="' + scriptname + '?page=' + tmp_page_encoded + '">' + tmp_page + '</a>&nbsp;</td>'
                        print '<td>' + tmp_host + '&nbsp;</td>'
                        print '<td>'
                        print strftime(datetime_fmt, time_tuple)
                        print ' &nbsp;</td></tr>'
                        if count > 20:
                            break
            print '</table>'
        else:
            print lang_dict['recentchanges_error']
        count = 0
        if rcs_info_available == 1 and not read_only(pagename):
            print '<p>' + lang_dict['checkout1'] + '<br>'
            for revision in page_revisions:
                count = count + 1
                print '<a href="' + scriptname + '?info=' + urllib.quote(pagename) + '&checkoutversion=' + urllib.quote(revision) + '">' + revision + '</a>, '
                if (count % 10) == 0:
                    print '<br>'
            if (count % 10) != 0:
                print '<br>'
            print lang_dict['checkout2'] + ' ' + pagename
            print '<br>' + lang_dict['checkout3']
            if form.has_key('checkoutversion'):
                check_out_release = urllib.unquote(form['checkoutversion'].value)
                if re.search('[0-9]+\.[0-9]+',check_out_release):
                    pagename = re.sub("'","",pagename)   # remove ' from pagename
                    check_out_release = re.sub("'","",check_out_release)   # remove ' from check_out_release
                    rcs_command = "cd " + data_dir + "; co -r'" + check_out_release + "' '" + pagename + "'\n"
                    p2 = os.popen(rcs_command)
                    print '<p>' + lang_dict['checkout4'] + ' ' + urllib.unquote(form['checkoutversion'].value) + ' ' + lang_dict['checkout5'] + '<p>'
                    print '<pre>'
                    input = p2.readlines()
                    for line in input:
                        print 'line = ' + line
                    print '</pre>'
    print '<hr>'
    print '<form method="post" action="%s">' % scriptname
    print '<a href="' + scriptname +'?search=' + urllib.quote(pagename) + '">' + lang_dict_main['search'] + '</a> ' + lang_dict_main['searchtext']
    print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<input type=submit value="' + lang_dict_main['goto'] + '"><input name=goto size=25 value="">'
    print '<br>'
    print '<a href="' + scriptname +'?index=' + urllib.quote(pagename) + '">' + lang_dict_main['index'] + '</a><br>'
    print '<a href="' + scriptname + '?page=' + urllib.quote(frontpage) +'">' + lang_dict_main['frontpage'] + '</a><br>'
    print '<a href="' + scriptname + '?page=admin">' + lang_dict_main['admin'] + '</a><br>'
    print '</form>'
    if left_menu == 'on':
        print_end_menu_table()
    print '</body></html>'
    
def get_raw_body(pagename):
    try:
        page = path.join(data_dir, pagename)
        foo = open(page, 'rt').read()
        return foo
    except IOError, er:
        if er.errno == errno.ENOENT:
            # just doesn't exist, use default
            return 'Describe %s here.' % pagename
        else:
            raise er

def editlog_add(page_name, host,error):
    # Write to the end of edit log file
    #
    editlog = open(editlog_name, 'a+')
    portalocker.lock(editlog, portalocker.LOCK_EX)
    try: 
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_EX)
        editlog.seek(0, 2)                  # to end
        editlog.write(error + ' ' + string.join((page_name, host, `time.time()`), "\t") + "\n")
    finally:
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()

def readlog_add(page_name, host):
    # log read access to a log file
    # write log as a web log
    tmp_date = time.time()
    time_tuple = localtime(float(tmp_date))
    readlog = open(readlog_name, 'a+')
    portalocker.lock(readlog, portalocker.LOCK_EX)
    file = data_dir + page_name
    foo = os.stat(file)
    file_size = re.sub('^([0-9]*).*',r'\1',repr(foo[6]))
    try:
        # Do a host name lookup
        hostname = gethostbyaddr(host)
    except:
        # probably DNS does not work
        hostname = [ host ]
    try:
        readlog.seek(0,2)
        if os.environ.has_key('HTTP_REFERER'):
            http_referer = os.environ['HTTP_REFERER']
        else:
            http_referer = '-'
        readlog.write(hostname[0] + ' - - [' + strftime('%d/%b/%Y:%k:%M:%S %Z',time_tuple) + '] "GET /kpwiki/' + urllib.quote(page_name) + ' ' + os.environ['SERVER_PROTOCOL'] + '" 200 ' + file_size + ' "' + http_referer + '" "' + os.environ['HTTP_USER_AGENT'] + '"\n')
        #readlog.write(string.join((page_name, host, `time.time()`), "\t") + "\n")
    finally:
        readlog.close()

def writelog_add(page_name, host,status):
    # log write access to a log file
    # write log as a web log
    tmp_date = time.time()
    time_tuple = localtime(float(tmp_date))
    if status == 'OLD':
        writelog = open(writelog_name, 'a+')
    elif status == 'NEW':
        writelog = open(writelog_new_name, 'a+')
    portalocker.lock(writelog, portalocker.LOCK_EX)
    file = data_dir + page_name
    foo = os.stat(file)
    file_size = re.sub('^([0-9]*).*',r'\1',repr(foo[6]))
    try:
        writelog.seek(0,2)
        writelog.write(gethostbyaddr(host)[0] + ' - - [' + strftime('%d/%b/%Y:%k:%M:%S %Z',time_tuple) + '] "PUT /kpwiki/' + urllib.quote(page_name) + ' ' + os.environ['SERVER_PROTOCOL'] + '" 200 ' + file_size + ' "-" "' + os.environ['HTTP_USER_AGENT'] + '"\n')
        #writelog.write(string.join((page_name, host, `time.time()`), "\t") + "\n")
    finally:
        writelog.close()

def tmp_filename(pagename):
    # Create a temp file name
    return path.join(data_dir, ('#' + pagename + '.' + `os.getpid()` + '#'))

def savepage(form,pagename,text,lang_dict_main,rcs_revision):
    # Save page to file
    #
    rcs_revision = rcs_revision + '\r\n'
    remote_name = environ.get('REMOTE_ADDR', '')
    tmpfilename = tmp_filename(pagename)
    lockfile = log_dir + pagename + '.LCK'
    if not re.search('\.\./',pagename) and not re.search('/',pagename) and not re.search('`',pagename):
        editlog_add(pagename, remote_name,'')
        f = open(tmpfilename, 'w')
        f.write(rcs_revision)
        f.write(text)
        f.close()
        page = data_dir + pagename
        rcs_command = "co -l '" + page + "'"
        os.system(rcs_command)
        if os.name == 'nt':
            # Bad Bill!  POSIX rename ought to replace. :-(
            try:
                os.remove(page)
            except OSError, er:
                if er.errno <> errno.ENOENT: raise er
        os.rename(tmpfilename, page)
        rcs_command = "ci -m'KpWiki edited page' -u '" + page + "'"
        os.system(rcs_command)
        writelog_add(pagename, remote_name, 'OLD')
        if form.has_key('edit_new_page'):
            writelog_add(pagename, remote_name,'NEW')
        if exists(lockfile):
            os.remove(lockfile)
        else:
            #print lang_dict_main['lock_removed']
            pass
    else:
        print '<h1>' + lang_dict_main['errortext'] + '</h1>'
        editlog_add(pagename, remote_name,'ERROR: trying to save a page with an invalid name')

# def edit_page()


def admin_default_lang_dict():
    # default language dictionary for admin page
    dict = {}
    dict['admin'] = 'Admin'
    dict['admin_header1'] = 'Admin pages and locks'
    dict['delete_page1'] = 'Delete a page'
    dict['delete_page2'] = '(press link and then select pages to delete)'
    dict['delete_page3'] = 'Delete page'
    dict['delete_page4'] = 'Page'
    dict['delete_page5'] = 'Delete'
    dict['delete_page6'] = 'Page deleted'
    dict['delete_page7'] = 'Page already deleted'
    dict['remove_lock1'] = 'Remove a lock on a page'
    dict['remove_lock2'] = '(press link and then select lock to remove)'
    dict['remove_lock3'] = 'No pages locked'
    dict['remove_lock4'] = 'Remove lock for page'
    dict['remove_lock5'] = 'Lock removed for page'
    dict['remove_lock6'] = 'Lock already removed for page'
    dict['undelete1'] = 'Undelete a page'
    dict['undelete2'] = '(press link and then select pages to undelete)'
    dict['undelete3'] = 'No pages deleted'
    dict['undelete4'] = 'Undelete page'
    dict['undelete5'] = 'Page'
    dict['undelete6'] = 'Undelete page'
    dict['undelete7'] = 'Undelete'
    dict['undelete8'] = 'Undeleted page'
    dict['upload1'] = 'Upload file'
    dict['upload2'] = 'Delete pictures'
    dict['upload3'] = 'File'
    dict['upload4'] = 'uploaded to server. You can now use'
    dict['upload5'] = 'ERROR: .. and / not allowed in filename'
    dict['upload6'] = 'ERROR: Could not save file'
    dict['upload7'] = 'Save as'
    dict['upload8'] = 'list directory contents of pictures and deleted pictures'
    dict['upload9'] = 'Picture name'
    dict['upload10'] = 'File name is missing'
    dict['upload11'] = 'Delete picture'
    dict['upload12'] = 'Submit delete'
    dict['upload13'] = 'or'
    dict['upload14'] = 'Picture already deleted'
    dict['upload15'] = 'Deleted picture'
    dict['upload16'] = 'There is no deleted pictures'
    dict['upload17'] = 'Undelete picture'
    dict['upload18'] = 'Picture'
    dict['upload19'] = 'Undelete pictures'
    dict['upload20'] = 'Undeleted picture'
    dict['upload21'] = 'Cannot create RCS directory. Wrong owner or permisson for image directory or directory is missing.'
    dict['upload22'] = 'Image directory does not exist'
    return dict


def admin_page(form,lang,lang_dict_main,server):
    # admin page
    # takes, form, language string, main language dictionary
    pagename = 'admin'
    lang_dict = admin_default_lang_dict()   # default language dictionary for edit
    lang_edit = lang_dir + lang + '.admin'
    if exists(lang_edit):
        lang_dict = read_lang_file(lang_edit, lang_dict)
    print '<head><title>KpWiki: ' + lang_dict['admin'] + '</title>'
    if css_url:
        print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
    print '</head><body>'
    if left_menu == 'on':
        wiki_page_content = []
        thispage = 'admin'
        print_menu_table(form, thispage, wiki_page_content, lang,lang_dict)
    print '<h1>'
    if left_menu != 'on':
        print logo_string
        print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
    print '<a href="' + scriptname + '?page=' + urllib.quote(pagename) + '">' + lang_dict['admin'] + '</a></h1>'
    print 'KpWiki ' +  lang_dict_main['version'] + ' ' + __version__ + ' ' + copyright + '<hr>'
    if form.has_key('Admin'):
        admin = urllib.unquote(form['Admin'].value)
        if admin == 'delete':
            # print a form and a table with all file names
            print '<form method="post" action="%s">' % scriptname
            files = os.listdir(data_dir)
            files.remove('RCS')  # Do not include RCS directory
            files = remove_readonly_pages(files)
            files = sort_ignorecase(files)
            print '<table border=1>'
            print '<tr><th>' + lang_dict['delete_page3'] + '</th><th>' + lang_dict['delete_page4'] + '</th></tr>'
            for page in files:
                print '<tr><td><input type=checkbox name=delete_' + urllib.quote(page) + '></td><td>' + page + '</td></tr>'
            print '</table>'
            print '<input type=submit value="' + lang_dict['delete_page5'] + '" name=deletepages>'
            print '<input type=hidden name="page" value="admin"'
            print '</form>'
        elif admin == 'remove lock':
            # show all file lock
            filelist = os.listdir(log_dir)
            # Exclude all files and directories except *.LCK
            files = []
            file_list = os.listdir(log_dir)
            for file in file_list:
                if re.search('.*\.LCK',file):
                    files.append(file)
            if len(files) != 0:
                for file in files:
                    print lang_dict['remove_lock4'] + ': <a href="' + scriptname + '?page=admin&unlock=' + urllib.quote(file) + '">' +  re.sub('(.*)\.LCK$',r'\1',file) + '</a><br>'
            else:
                print lang_dict['remove_lock3'] + '<br>'
        elif admin == 'undelete':
            # print a form and a table with all deleted files
            print '<form method="post" action="%s">' % scriptname
            rcs_dir = data_dir + 'RCS'
            deleted_files = []
            files = os.listdir(rcs_dir)
            files = sort_ignorecase(files)
            for file in files:
                file = re.sub('(.*),v',r'\1',file)
                if not exists(data_dir + file):
                    deleted_files.append(file)
            if len(deleted_files) == 0:
                print lang_dict['undelete3'] + '<p>'
            else:
                print '<table border=1>'
                print '<tr><th>' + lang_dict['undelete4'] + '</th><th>' + lang_dict['undelete5'] + '</th></tr>'
                for file in deleted_files:
                    print '<tr><td><input type=checkbox name=undelete_' + urllib.quote(file) + '></td><td>' + file + '</td></tr>'
                print '</table>'
                print '<input type=submit value="' + lang_dict['undelete7'] + '" name=undeletepages>'
            print '<input type=hidden name="page" value="admin"'
            print '</form>'
        else:
            print 'Error: '
            print 'Que?'
    elif form.has_key('deletepages'):
        # DELETE pages
        for key in form.keys():
            if re.search('^delete_',key):
                key = urllib.unquote(key)
                if re.search('\.\.',key) or re.search('/',key):
                    key = ''
                else:
                    key = re.sub('^delete_(.*)',r'\1',key)
                if not read_only(key):
                    try:
                        page = data_dir + key
                        os.remove(page)
                        print lang_dict['delete_page6'] + ': ' + key
                    except OSError:
                        print lang_dict['delete_page7'] + ': ' + key + '<br>'
            print '<br>'
    elif form.has_key('undeletepages'):
        # UNDELETE pages
        for key in form.keys():
            if re.search('^undelete_',key):
                key = urllib.unquote(key)
                key = re.sub('^undelete_(.*)',r'\1',key)
                if not (re.search('\.\.',key) or re.search('/',key)):
                    key = re.sub('^delete_(.*)',r'\1',key)
                    key = re.sub("'","",key)   # remove ' from key
                    rcs_command = "cd " + data_dir + "; co '" + key + "'"
                    p = os.popen(rcs_command)
                    print lang_dict['undelete8'] + ': ' + key
    elif form.has_key('unlock'):
        file = urllib.unquote(form['unlock'].value)
        if re.search('\.\.',file) or re.search('/',file):
            file=''
        else:
            lockfile = log_dir + file
        try:
            os.remove(lockfile)
            print lang_dict['remove_lock5'] + ': ' +  re.sub('(.*)\.LCK$',r'\1',file) + '<p>'
        except OSError:
            print lang_dict['remove_lock6'] + ': ' + re.sub('(.*)\.LCK$',r'\1',file) + '<p>'
    else:
        print '<h2>' + lang_dict['admin_header1'] + '</h2>'
        print '<a href="' + scriptname + '?page=admin&Admin=delete">' + lang_dict['delete_page1'] + '</a> ' + lang_dict['delete_page2'] + '<p>'
        print '<a href="' + scriptname + '?page=admin&Admin=remove%20lock">' + lang_dict['remove_lock1'] + '</a> ' + lang_dict['remove_lock2'] + '<p>'
        print '<a href="' + scriptname + '?page=admin&Admin=undelete">' + lang_dict['undelete1'] + '</a> ' + lang_dict['undelete2'] + '<p>'
    print '<FORM METHOD=POST ENCTYPE="multipart/form-data" ACTION="'+ scriptname + '">'
    print '<hr>' + lang_dict['upload1'] + ': <INPUT TYPE=FILE NAME="upfile">'
    print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;' + lang_dict['upload7'] + ' <input name="save_upload_as" size=20 value=""> '
    print '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<INPUT TYPE=SUBMIT VALUE="' + lang_dict['upload2'] + '">'
    print '<input type=hidden name="page" value="admin">'
    print '</FORM>'    
    if form.has_key('upfile'):
        print '<p>'
        file_name_ok = 1
        if form.has_key('save_upload_as') and form['save_upload_as'].value != '':
            upload_filename = urllib.unquote(form['save_upload_as'].value)
        elif form['upfile'].value != '':
            upload_filename = urllib.unquote(form['upfile'].filename)
        else:
            file_name_ok = 0
            upload_filename = ''
        if re.search('\.\.',upload_filename) or re.search('/',upload_filename) or re.search('`',upload_filename):
            print lang_dict['upload5'] + '<p>'
        else:
            try:
                count = 1
                new_filename = upload_filename
                upload_file_dir = web_tree + image_dir + new_filename
                while exists(upload_file_dir):
                    new_filename = repr(count) + "-" + upload_filename
                    upload_file_dir = web_tree + image_dir + new_filename
                    count = count + 1
                print '<p>'
                upload = open(upload_file_dir, 'w')
                upload.write(form['upfile'].value)
                upload.close()
                rcs_command = "cd " + web_tree + image_dir + "; ci -m'KpWiki saved the picture file' -u '" + new_filename + "'"
                os.system(rcs_command)
                print lang_dict['upload3'] + ' ' + upload_filename + ' ' + lang_dict['upload4'] + '<br><b>[img-link:' + server + image_dir +  new_filename + ']</b>'
                print '<br>' + lang_dict['upload13'] + '<br><b>[img:' + new_filename + ']</b><p>'
            except IOError:
                print lang_dict['upload6'] + ': '                
                if file_name_ok == 0:
                    print '<b>&lt;' + lang_dict['upload10'] + '&gt;</b>'
                else:
                    print upload_file_dir
    if form.has_key('submit_delete_picture'):
        # delete picture
        for key in form.keys():
            if re.search('^delete_picture_',key):
                key = urllib.unquote(key)
                key = re.sub('^delete_picture_(.*)',r'\1',key)
                try:
                    picture = web_tree + image_dir + key
                    os.remove(picture)
                    print lang_dict['upload15'] + ': ' + key
                except OSError:
                    print lang_dict['upload14'] + ': ' + key + '<br>'
            print '<br>'
    if form.has_key('submit_undelete_picture'):
        for key in form.keys():
            if re.search('^undelete_picture_',key):
                key = urllib.unquote(key)
                key = re.sub('^undelete_picture_(.*)',r'\1',key)
                key = re.sub("'","",key)   # remove ' from key
                rcs_command = "cd " + web_tree + image_dir  + "; co '" + key + "'"
                p = os.popen(rcs_command)
                print lang_dict['upload20'] + ': ' + key + '<br>'
    print '<FORM METHOD=POST ENCTYPE="multipart/form-data" ACTION="'+ scriptname + '">'
    print '<p><input type=submit name=upload_dir_browse value="'+ lang_dict['upload8'] + '">'
    if form.has_key('upload_dir_browse'):
        print '<p>'
        print '<form method="post" action="%s">' % scriptname
        print '<table border=1><tr><td align=center>'
        print '<table border=1 bgcolor=#aadede ><tr><th>' + lang_dict['upload11'] + '</th><th>' + lang_dict['upload9'] + '</th></tr>'
        if exists(web_tree + image_dir):
            filelist = os.listdir(web_tree + image_dir)
            filelist.sort()
        else:
            print lang_dict['upload22'] + '<p>'
            filelist = []
        if filelist.count('RCS') > 0:
            filelist.remove('RCS')
        for file in filelist:
            print '<tr><td><input type=checkbox name=delete_picture_' + urllib.quote(file) + '></td>'
            print '<td>' + file + '</td></tr>'
        print '</table></td><td>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td align=center>'
        rcs_dir = web_tree + image_dir + 'RCS'
        deleted_files = []
        if exists(rcs_dir):
            files = os.listdir(rcs_dir)
            files = sort_ignorecase(files)
        else:
            files = []
            command = 'mkdir ' + rcs_dir
            p = os.popen(command)
            if not exists(rcs_dir):
                print lang_dict['upload21'] + '<p>'
        for file in files:
            file = re.sub('(.*),v',r'\1',file)
            if not exists(web_tree + image_dir + file):
                deleted_files.append(file)
        if len(deleted_files) == 0:
            print lang_dict['upload16'] + '<p>'
        else:
            print '<table border=1>'
            print '<tr><th>' + lang_dict['upload17'] + '</th><th>' + lang_dict['upload18'] + '</th></tr>'
            for file in deleted_files:
                print '<tr><td><input type=checkbox name=undelete_picture_' + urllib.quote(file) + '></td><td>' + file + '</td></tr>'
            print '</table>'
        print '</td></tr>'
        print '<tr><td align=center>'
        print '<input type=submit name=submit_delete_picture value="' + lang_dict['upload12'] + '"></td><td>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td align=center>'
        print ' <input type=submit name=submit_undelete_picture value="' + lang_dict['upload19'] + '">'
        print '</td></tr>'
        print '</table>'
        print '</form>'
    print '<input type=hidden name="page" value="admin">'
    print '</FORM>'
    print '<hr>'
    print '<form method="post" action="%s">' % scriptname
    print '<a href="' + scriptname +'?search=' + urllib.quote(pagename) + '">' + lang_dict_main['search'] + '</a> ' + lang_dict_main['searchtext']
    print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<input type=submit value="' + lang_dict_main['goto'] + '"><input name=goto size=25 value="">'
    print '<br>'
    print '<a href="' + scriptname +'?index=' + urllib.quote(pagename) + '">' + lang_dict_main['index'] + '</a><br>'
    print '<a href="' + scriptname + '?page=' + urllib.quote(frontpage) +'">' + lang_dict_main['frontpage'] + '</a>'
    print '</form>'
    if left_menu == 'on':
        print_end_menu_table()
    print '</body></html>'
    
def main_default_lang_dict():
    # Default texts for main
    dict={}
    dict['version'] = 'version'
    dict['last_mod'] = 'Last modified'
    dict['EditText'] = 'EditText'
    dict['edittext'] = '(edit this page.)'
    dict['search'] = 'Search'
    dict['searchtext'] = '(search for pages including text.)'
    dict['goto'] = 'Goto page'
    dict['index'] = 'Index'
    dict['wordindex'] = 'Word index'
    dict['frontpage'] = 'Go to frontpage'
    dict['errortext'] = 'ERROR !!! CAN NOT SAVE THIS PAGE !!! YOU CAN NOT HAVE ../ OR / IN PAGE NAME'
    dict['info_about_page'] = 'More information about this page'
    dict['info_error'] = 'No page found'
    dict['no_rcs_info'] = 'No rcs information available'
    dict['rcs_diff_help'] = 'Press revision link above for a diff between that revision and the latest revision'
    dict['rcs_diff_text'] = 'Diff betweeen following revisions of the page'
    dict['lock_removed'] = 'Lockfile removed by user'
    dict['page_revision'] = 'Page revision'
    dict['admin'] = 'Go to admin page'
    return dict

def log_read_count(pagename):
    # count number of read access for a file/KpWiki page
    count = 1
    countfilename = log_dir + 'count/' + pagename
    if exists(countfilename):
        # write count + 1 to file
        file = open(countfilename, 'r+')
        portalocker.lock(file, portalocker.LOCK_EX)
        count = string.atoi(file.read())
        count = count + 1
        file.seek(0)
        file.write(repr(count))
        file.write('\n')
        file.close()
    else:
        # write 1 to the new file
        file = open(countfilename, 'w')
        portalocker.lock(file, portalocker.LOCK_EX)
        # write 1
        file.write('1\n')
        file.close()

def main(lang,charset):
    lang_dict = main_default_lang_dict()   # default language dictionary
    lang_main = lang_dir + lang + '.main'
    if exists(lang_main):
        lang_dict = read_lang_file(lang_main, lang_dict)
    else:
        lang=''
    page = ""
    showpage = 0
    cancel = 0
    emit_header(charset)
    try:
        form = cgi.FieldStorage()
        remote_name = environ.get('REMOTE_ADDR', '')
        server = environ.get('HTTP_HOST')
        if form.has_key('edit_lock_override'):
            cancel = 1
            # cancel editing page
            if form.has_key('edit_lock_file'):
                if exists(urllib.unquote(form['edit_lock_file'].value)):
                    delete_file = urllib.unquote(form['edit_lock_file'].value)
                    if not (re.search('\.\.',delete_file)):
                        os.remove(delete_file)
        if form.has_key('edit'):
            # edit this page
            edit_page_name = urllib.unquote(form['edit'].value)
            if not (re.search('\.\.',edit_page_name) or re.search('/',edit_page_name)):
                edit_page(form,edit_page_name,lang,lang_dict)
        elif form.has_key('index'):
            # index page
            makeindex(form,'index',lang,lang_dict)
        elif form.has_key('info'):
            pagename = urllib.unquote(form['info'].value)
            if not (re.search('\.\.',pagename) or re.search('/',pagename)):
                info_about_page(form,lang,pagename,lang_dict)
        elif form.has_key('goto'):
            pagename = urllib.unquote(form['goto'].value)
            if not (re.search('\.\.',pagename) or re.search('/',pagename)):
                if pagename == 'index':
                    makeindex(form,'index',lang,lang_dict)
                elif pagename == 'search':
                    search(form,lang,lang_dict)
                elif pagename == 'Search':
                    search(form,lang,lang_dict)
                elif pagename == 'Admin' or pagename == 'admin':
                    # goto admin page
                    admin_page(form,lang,lang_dict,server)
                else:
                    showpage = 1
        elif form.has_key('search'):
            # search
            search(form,lang,lang_dict)
        elif form.has_key('page'):
            # go to a new page
            pagename = urllib.unquote(form['page'].value)
            if not (re.search('\.\.',pagename) or re.search('/',pagename)):
                if pagename == 'index':
                    # goto index page
                    makeindex(form,'index',lang,lang_dict)
                elif pagename == 'search':
                    # goto search page
                    search(form,lang,lang_dict)
                elif pagename == 'Admin' or pagename == 'admin':
                    # goto admin page
                    admin_page(form,lang,lang_dict,server)
                else:
                    showpage = 1
        else:
            # go to frontpage
            pagename = frontpage
            showpage = 1
        if showpage == 1:
            if form.has_key('savepage') and cancel == 0:
                rcs_revision = '$Revision: 1.0 $'
                if form.has_key('rcs_revision'):
                    rcs_revision = urllib.unquote(form['rcs_revision'].value)
                    if not re.search('\$Revision:[ 	]+[0-9]+\.[0-9]+[ 	]+\$',rcs_revision):
                        rcs_revision = '$Revision: 1.0 $'
                pagename = urllib.unquote(form['savepage'].value)
                if form.has_key('savetext'):
                    #savetext = urllib.unquote(form['savetext'].value)
		    savetext = form['savetext'].value
                else:
                    savetext = ''
                if pagename != 'index' and not (re.search('\.\.',pagename) or re.search('/',pagename) or re.search('`',pagename)):
                    savepage(form,pagename,savetext,lang_dict,rcs_revision)
                if pagename == 'left_menu':
                    if form.has_key('old_page'):
                        pagename = frontpage
            page = path.join(data_dir, pagename)
            if exists(page):
                infile = open(page,'r')
                log_read_count(pagename)
                readlog_add(pagename,remote_name)
                input = infile.readlines()
                wiki_page_content = input
                rcs_revision = ''
                if re.search('\$Revision:',input[0]):
                    rcs_revision = re.sub('\$Revision:[ ]*([^\$]*)\$',r'\1',input[0])
                    del input[0]
                input = wiki_to_html(input,server)
                print '<head><title>KpWiki: ' + pagename + '</title>'
                if css_url:
                    print '<link rel="stylesheet" type="text/css" href="%s">' % css_url
                print '</head><body>'
                if left_menu == 'on':
                    print_menu_table(form, pagename, wiki_page_content, lang,lang_dict)
                print '<h1>'
                if left_menu != 'on':
                    print logo_string
                    print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'
                print '<a href="' + scriptname + '?page=' + pagename + '">' + pagename + '</a></h1>'
                print 'KpWiki ' +  lang_dict['version'] + ' ' + __version__ + ' ' + copyright + '<hr>'
                for i in range(len(input)):
                    print input[i]
                print '<p><hr>'
                # Get page version
                rcs_version = ''
                if exists(data_dir + 'RCS' + pagename):
                    pagename = re.sub("'","",pagename)   # remove ' from pagename
                    rcs_command = "cd " + data_dir + "; rlog -h '" + pagename + "'"
                    p = os.popen(rcs_command)
                    rcs_input = p.readlines()
                    for line in rcs_input:
                        rcs_version = ''
                print '<br>' + lang_dict['last_mod'] + ': ' + mod_time(pagename) + ' &nbsp;&nbsp;&nbsp;' + ' &nbsp;&nbsp;'  + lang_dict['page_revision'] + ': ' +  rcs_revision
                print '<p>'
                print '<form method="post" action="%s">' % (scriptname)
                old_page = urllib.quote(scriptname) + '?page=' + urllib.quote(pagename)
                print '<input type=hidden name="old_page" value="' + old_page + '">'
                if not read_only(pagename):
                    print '<a href="' + scriptname + '?edit=' + urllib.quote(pagename) + '">' + lang_dict['EditText'] + '</a> ' + lang_dict['edittext'] + '<br>'
                print '<a href="' + scriptname + '?search=' + urllib.quote(pagename) + '">' + lang_dict['search'] + '</a> ' + lang_dict['searchtext']
                print ' &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<input type=submit value="' + lang_dict['goto'] + '"><input name=goto size=25 value="">'
                print '<br>'
                print '<a href="' + scriptname + '?index=' + urllib.quote(pagename) + '">' + lang_dict['index'] + '</a><br>'
                print '<a href="' + scriptname + '?page=' + urllib.quote(frontpage) +'">' + lang_dict['frontpage'] + '</a><br>'
                print '<a href="' + scriptname + '?page=admin">' + lang_dict['admin'] + '</a><br>'
                print '</form>'
                if left_menu == 'on':
                    print_end_menu_table()
                print '</body></html>'
            elif cancel == 1 and form.has_key('old_page'):
                print '<HEAD>'
                print ' <META HTTP-EQUIV=REFRESH CONTENT="1;URL=' + form['old_page'].value + '">'
                print '</HEAD>'
            else:
                # just doesn't exist, use default
                edit_page(form, pagename,lang,lang_dict)
                # <META HTTP-EQUIV=REFRESH CONTENT="5;URL=http://ntt.com/">
    except:
         cgi.print_exception()
    sys.stdout.flush()


# Configurable parts ------------------------------------------

#sys.stderr = open('/tmp/mbp_kpwiki_err', 'at')

# KpWiki CONFIGURATION FILE
kpwiki_conf = '/opt/kpwiki/etc/kpwiki.conf'
kpwiki_readonly_pages = '/opt/kpwiki/etc/read-only-pages'   # Read only wiki pages

# Default values
kpwiki_dir = '/opt/kpwiki/'
errorfile = kpwiki_dir + 'log/mbp_kpwiki_err'
sys.stderr = open(errorfile, 'at')
web_tree = '/var/www/html/'
logo_string = '<img src="/pictures/kpwiki.gif" border=0 alt="kpwiki">'
menu_logo_string = '<img src="/pictures/kpwiki-transparent.gif" border=0 alt="kpwiki">'
icon_dir = '/pictures'
image_dir = '/kpwiki-images/'
changed_time_fmt = ' [%I:%M %p]'
date_fmt = '%a %d %b %Y'
datetime_fmt = '%a %d %b %Y %I:%M %p'
show_hosts = '0'                          # show hostnames?
css_url = '/css/kpwiki.css'         # stylesheet link, or ''
nonexist_qm = '0'                         # show '?' for nonexistent?
show_no_lines_on_hit = '7'                # number of lines to show on a search hit
left_menu = 'on'                          # show a menu at left side
menu_width = '17'                       # menu width in %
frontpage = 'Startsida'                 # start page, front page
charset = 'iso-latin1'
site_lock = '0'                         # lock entire site. No pages can be updated

if exists(kpwiki_conf):
    f = open(kpwiki_conf)
    lines=f.readlines()
    dict = {}

    for line in lines:
        line = re.sub('^[ ]*#.*\r*\n*',r'',line)
        if line != '' and line != '\n' and line != '\r':
            var = re.sub('([^ =]*)([ =].*)[\n\r]*',r'\1',line)
            value = re.sub('([^ =]*)[ =]*(.*)[\n\r]*',r'\2',line)
            dict[var] = value

    if dict.has_key('kpwiki_dir'):
        kpwiki_dir = dict['kpwiki_dir']

    errorfile = kpwiki_dir + 'log/mbp_kpwiki_err'
    sys.stderr = open(errorfile, 'at')


    if dict.has_key('logo_string'):
        logo_string = dict['logo_string']

    if dict.has_key('web_tree'):
        web_tree = dict['web_tree']

    if dict.has_key('icon_dir'):
        icon_dir = dict['icon_dir']

    if dict.has_key('image_dir'):
        # upload images
        image_dir = dict['image_dir']

    if dict.has_key('changed_time_fmt'):
        changed_time_fmt = ' ' + dict['changed_time_fmt']

    if dict.has_key('date_fmt'):
        date_fmt = dict['date_fmt']

    if dict.has_key('datetime_fmt'):
        datetime_fmt = dict['datetime_fmt']

    if dict.has_key('show_hosts'):
        show_hosts = dict['show_hosts']

    if dict.has_key('css_url'):
        css_url = dict['css_url']

    if dict.has_key('nonexist_qm'):
        nonexist_qm = dict['nonexist_qm']

    if dict.has_key('show_no_lines_on_hit'):
        show_no_lines_on_hit = dict['show_no_lines_on_hit']

    if dict.has_key('frontpage'):
        frontpage = dict['frontpage']

    if dict.has_key('lang'):
        lang = dict['lang']

    if dict.has_key('charset'):
        charset = dict['charset']
        
    if dict.has_key('left_menu'):
        left_menu = dict['left_menu']

    if dict.has_key('menu_width'):
        menu_width = dict['menu_width']

    if dict.has_key('menu_logo_string'):
        menu_logo_string = dict['menu_logo_string']

    if dict.has_key('site_lock'):
        site_lock = dict['site_lock']

else:
    pass

# DO NOT CHANGE copyright.
# Before you do any changes to KpWiki make sure you understand GNU GENERAL PUBLIC LICENSE Version 2
copyright = '&nbsp;&nbsp;&nbsp;Copyright (C) 2003 Kjell Enblom.<br>GNU GENERAL PUBLIC LICENSE Version 2. KpWiki comes with ABSOLUTELY NO WARRANTY'

sys.path.append(kpwiki_dir)
import portalocker
data_dir = path.join(kpwiki_dir, 'data/')
lang_dir = path.join(kpwiki_dir, 'lang/')
log_dir = path.join(kpwiki_dir, 'log/')
editlog_name = path.join(log_dir, 'editlog')
readlog_name = path.join(log_dir, 'access_log')
writelog_name = path.join(log_dir, 'write_log')
writelog_new_name = path.join(log_dir, 'write_new_log')
cgi.logfile = path.join(data_dir, 'cgi_log')
scriptname= get_scriptname()
main(lang,charset)
